# Copyright (c) Regents of the University of Minnesota
# Copyright (c) Michael Gilbert
# Distributed under the terms of the Modified BSD License.

"""Batch spawners

This file contains an abstraction layer for batch job queueing systems, and implements
Jupyterhub spawners for Torque, SLURM, and eventually others.

Common attributes of batch submission / resource manager environments will include notions of:
  * queue names, resource manager addresses
  * resource limits including runtime, number of processes, memory
  * singleuser child process running on (usually remote) host not known until runtime
  * job submission and monitoring via resource manager utilities
  * remote execution via submission of templated scripts
  * job names instead of PIDs
"""
import asyncio
from concurrent.futures import ThreadPoolExecutor
import os
import pprint
import pwd
import re
import signal
from subprocess import Popen, call
import subprocess
import sys, traceback
from urllib.parse import quote, urlparse

from tornado import gen
from tornado.iostream import StreamClosedError
from tornado.process import Subprocess
from traitlets import (
    Instance, Integer, Unicode, Float, Dict
)

from kslhub.ssh_authenticator import session_manager
from jupyterhub import dbutil, orm
from jupyterhub.app import JupyterHub
from jupyterhub.spawner import Spawner
from jupyterhub.spawner import set_user_setuid
from jupyterhub.utils import random_port
import xml.etree.ElementTree as ET

from kslhub import scan_template
import glob
import re

# add on to compute secret and token to be passed to the spawning job
# import os
# from jupyterhub.services.auth import HubAuth
# auth = HubAuth(
#     api_token=os.environ['JUPYTERHUB_API_TOKEN'],
#     cookie_cache_max_age=60,
# )
kslhub_root = "."  # os.getenv('KSLHUB_ROOT')

slurm_machine = session_manager()

@gen.coroutine
def run_command(cmd, input=None, env=None, user=None):
    print("[AUTHENT] cmd run : |||%s||| " % cmd)
    #print("SKKKKKKKKKk env : |||%s||| " % pprint.pformat(env))
    #print("SKKKKKKKKKk cmd : |||%s||| " % pprint.pformat(cmd))
    #input_str = "|||%s||| " % pprint.pformat(input)
    #print("SKKKKKKKKKk input : %s, find_result: %s" % (input_str, input_str.find("|||None|||")))
    #print("[AUTHENT]================ here user=%s!!!" % user)
    #print("[AUTHENT] slurm_machine=%s   or %s" % (slurm_machine,pprint.pformat(slurm_machine)))
    if user==None:
        print("[AUTHENT]===============  here user=None!!!")
        traceback.print_stack()
        print("[AUTHENT]===============  here user=None!!!")
    try:
        connection = slurm_machine.get_session(user)
        if not(connection):
            print("[AUTHENT]==================== session expired for %s/%s " %\
                  (slurm_machine.machine,user))
            traceback.print_stack()
            print("[AUTHENT]==================== session expired for %s/%s " %\
                  (slurm_machine.machine,user))
            return False
            # test de la connection
            return redirect(auth.login_url + '?next=%s' % quote(request.path))
  
    except Exception as e:
        print("[AUTHENT] no connection: problem in getting the session")
        print(e)
        traceback.print_exc(file=sys.stdout)
        return "problem starting ssh session"
        
    stdin, stdout, stderr = connection.exec_command(cmd)
    if not(input==None):
        print("sending input:/%s/" % input)
        stdin.write("".join(input))
        stdin.channel.shutdown_write()
    else:
        print("[AUTHENT] no input")
 
        
    output = stdout.readlines()
    print("[AUTHENT] output for cmd=|||%s||| = %s " %    (cmd,output))
    try:
        res = output[-1][:-1]
    except:
        print("[AUTHENT] problem in parsing output. forcing res at nothing")
        res = ""
    print("[AUTHENT] res for cmd=|||%s||| = %s " %    (cmd,res))
    return res


class BatchSpawnerBase(Spawner):
    """Base class for spawners using resource manager batch job submission mechanisms

    This base class is developed targetting the TorqueSpawner and SlurmSpawner, so by default
    assumes a qsub-like command that reads a script from its stdin for starting jobs,
    a qstat-like command that outputs some data that can be parsed to check if the job is running
    and on what remote node, and a qdel-like command to cancel a job. The goal is to be
    sufficiently general that a broad range of systems can be supported with minimal overrides.

    At minimum, subclasses should provide reasonable defaults for the traits:
        batch_script
        batch_submit_cmd
        batch_query_cmd
        batch_cancel_cmd

    and must provide implementations for the methods:
        state_ispending
        state_isrunning
        state_gethost
    """

    # override default since batch systems typically need longer
    start_timeout = Integer(300, config=True)

    # override default server ip since batch jobs normally running remotely
    ip = Unicode("0.0.0.0", config=True, help="Address for singleuser server to listen at")

    # all these req_foo traits will be available as substvars for templated strings
    req_queue = Unicode('', config=True, \
        help="Queue name to submit job to resource manager"
        )

    req_host = Unicode('', config=True, \
        help="Host name of batch server to submit job to resource manager"
        )

    req_memory = Unicode('', config=True, \
        help="Memory to request from resource manager"
        )

    req_nprocs = Unicode('', config=True, \
        help="Number of processors to request from resource manager"
        )

    req_runtime = Unicode('', config=True, \
        help="Length of time for submitted job to run"
        )

    req_options = Unicode('', config=True, \
        help="Other options to include into job submission script"
        )

    hub_port = Integer(8080, config=True, \
        help="Port where the hub is talking"
        )

    hub_ip = Unicode('localhost', config=True, \
        help="IP of the hub"
        )

    proxy_port = Integer(8081, config=True, \
        help="Port where the proxy is talking"
        )

    proxy_ip = Unicode('localhost', config=True, \
        help="IP of the proxy"
        )

    req_username = Unicode()
    def _req_username_default(self):
        return self.user.name

    # Useful IF getpwnam on submit host returns correct info for exec host
    req_homedir = Unicode()
    def _req_homedir_default(self):
        return pwd.getpwnam(self.user.name).pw_dir

    req_keepvars = Unicode()
    def _req_keepvars_default(self):
        return ','.join(self.get_env().keys())

    batch_script = Unicode('', config=True, \
        help="Template for job submission script. Traits on this class named like req_xyz "
             "will be substituted in the template for {xyz} using string.Formatter. "
             "Must include {cmd} which will be replaced with the jupyterhub-singleuser command line."
        )

    # Raw output of job submission command unless overridden
    job_id = Unicode()

    # Will get the raw output of the job status command unless overridden
    job_status = Unicode()

    # Prepare substitution variables for templates using req_xyz traits
    def get_req_subvars(self):
        names = self.trait_names()
        reqlist = [ t for t in self.trait_names() if t.startswith('req_') ]
        # need to add req_input manually..
        for k,v in self.__dict__.items():
            if k[:4] == "dyn_":
               reqlist = reqlist + [k]
        subvars = {}
        for t in reqlist:
            subvars[t[4:]] = getattr(self, t)
        return subvars

    batch_submit_cmd = Unicode('', config=True, \
        help="Command to run to submit batch scripts. Formatted using req_xyz traits as {xyz}."
        )

    def parse_job_id(self, output):
        "Parse output of submit command to get job id."
        return output

    def cmd_formatted_for_batch(self):
        return ' '.join(self.cmd + self.get_args())

    @gen.coroutine
    def submit_batch_script(self):
        subvars = self.get_req_subvars()
        cmd = self.batch_submit_cmd.format(**subvars)
        self.log.info("\nsubvars:%s\n" % pprint.pformat(subvars))
        
        for v in subvars['keepvars'].split(","):
            if os.getenv(v):
                subvars[v] = os.getenv(v)

        self.log.info("\nsubvars after keepvar :%s\n" % pprint.pformat(subvars))
        
        #SKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKkkk
        template_directory = '%s/' % self.authenticator.job_template_dir
        template_to_find =  "%s/%s -*.template" % (template_directory,subvars['input'])
        self.log.info('template_to_find: /%s/' % template_to_find)

        templates = glob.glob(template_to_find)

        template = templates[0]

        self.log.info("template found %s" % template)

        script = "".join(open(template,"r").readlines())
        
        self.log.info('template content: \n' + '='*80 + '\n' + script + '\n' + '='*80 + '\n')

        # replace all the __tag;...;..;;...__ and __tag__ pattern with the value
        # chosen in the GUI
        
        for k in subvars.keys():
            v = subvars[k]
            script = re.sub(r"__%s;.+__" % k , v, script)
            script = re.sub(r"__%s__" % k , v, script)
            
        self.log.info('content after __tag__ replacememnt: \n' \
                      + '='*80 + '\n' + script + '\n' + '='*80 + '\n')
        
        self.log.info('user_options: %s' % self.user_options)


        # initialize the keepvar environment variable

        # find the right spot in the script (just before the first command

        script_env_set = ""
        env_done = False

        for l in script.split("\n"):
            if len(l)==0:
                script_env_set = script_env_set + l + "\n"
            elif l[0]=="#" or len(l.strip())==0 or env_done:
                script_env_set = script_env_set + l + "\n"
            else:
                vars = "#" * 80 + \
                       "\n# Script generated from template\n#\t%s" % os.path.abspath(template) + \
                       "\n" +"#" * 80 + \
                       "\n# Environment configuration to be forwarded...." +\
                       "\n" +"#" * 80 + "\n\n"
                for v in subvars['keepvars'].split(","):
                    if os.getenv(v):
                        vars = vars + "export %s=%s\n" % (v,os.getenv(v))
                vars = vars + \
                       "export USER=__USER__" + \
                       "\nexport KSLHUB_ROOT=__KSLHUB_ROOT__" +\
                       "\n. $KSLHUB_ROOT/kslhub_init_env.sh" +\
                       "\n\n" +\
                       "#" * 80 + "\n# Setting connection with the hub\n" +\
                       "#" * 80
                
                script_env_set = script_env_set + vars

                # adding token and secret credential

                script_env_set = script_env_set + \
                "\nexport JUPYTERHUB_BASE_URL=/" +\
                "\nexport JUPYTERHUB_CLIENT_ID=jupyterhub-user-__USER__" +\
                "\nexport JUPYTERHUB_API_TOKEN=__TOKEN__" +\
                "\nexport JUPYTERHUB_API_URL=http://__HOST__:__PORT__/hub/api" +\
                "\nexport JUPYTERHUB_USER=__USER__" +\
                "\nexport JUPYTERHUB_OAUTH_CALLBACK_URL=/user/__USER__/oauth_callback" +\
                "\nexport JUPYTERHUB_HOST=" +\
                "\nexport JUPYTERHUB_SERVICE_PREFIX=/user/__USER__/" +\
                "\nexport JUPYTER_RUNTIME_DIR=__KSLHUB_ROOT__/runtime" +\
                "\nexport CONFIGPROXY_AUTH_TOKEN=__SECRET__" +\
                "\n\n"+ "#" * 80 + "\n" +\
                "# Original script" +\
                "\n"+ "#" * 80 + "\n" +\
                l + "\n"
                
                env_done = True
                

        script = script_env_set
        self.log.info('content after env addition: \n' \
                      + '='*80 + '\n' + script + '\n' + '='*80 + '\n')


        
        # computing the token related to user
        
        hub = JupyterHub(parent=self)
        hub.load_config_file(hub.config_file)
        hub.init_db()
        def init_users():
            loop = asyncio.new_event_loop()
            loop.run_until_complete(hub.init_users())
        ThreadPoolExecutor(1).submit(init_users).result()
        user = orm.User.find(hub.db, self.user.name)
        self.log.info("user : %s" % pprint.pformat(self.user))
        if user is None:
            self.log.info("No such user: %s" % self.user.name, file=sys.stderr)
            self.exit(1)
        token = user.new_api_token(note="command-line generated")
        print(token)
        
        self.log.info("token = !!!%s!!! " % token)

        script = script.replace("__TOKEN__",token)


        client_id = 'jupyterhub-user-%s' % quote(self.user.name)
        oauth_client = (
            hub.db
            .query(orm.OAuthClient)
            .filter_by(identifier=client_id)
            .first()
        )
        self.log.info("secret = !!!%s!!! " % oauth_client.secret)

        script = script.replace("__SECRET__", oauth_client.secret)
        script = script.replace("__USER__", self.user.name)
        # SKKKKKKKK to be fixed!!!
        script = script.replace("__HOST__", self.proxy_ip)
        script = script.replace("__PORT__", str(self.proxy_port))
        script = script.replace("__NOTEBOOK_PORT__",str(self.user.server.port))
        script = script.replace("__PYTHONPATH__", os.getenv("PYTHONPATH",""))
        script = script.replace("__KSLHUB_ROOT__", os.getenv("KSLHUB_ROOT","."))
        script = script.replace("__USER__", os.getenv("USER","$USER"))
        script = script.replace("__JOB_DIR__", os.getenv("KSLHUB_ROOT",".")+"/jobs/")
        

        self.log.info('content after token replacement: \n' \
                      + '='*80 + '\n' + script + '\n' + '='*80 + '\n')

        
        
        subvars['cmd'] = self.cmd_formatted_for_batch()
        self.log.info("cmd formatted:%s" % subvars['cmd'])
        script = script.replace("__CONNECT__", subvars['cmd'])
        
        if hasattr(self, 'user_options'):
            subvars['user_options'] = self.user_options

        #script = self.batch_script.format(**subvars)
        #self.log.debug('Spawner submitting job using ' + cmd)
        #self.log.debug('Spawner submitted script:\n' + script)


        # set all 
        
            
        f = open("%s/jobs/%s_last_script" %  (kslhub_root,self.user.name),"w")
        f.write(script)
        f.close()



        if script.find("#SBATCH -v")>-1:
            # need to export environment variable by dynamically computing their absolute values
            vars = "#" * 80 + "\n# Environment variable to be forwarded....\n"
            for v in subvars['keepvars'].split(","):
                if os.getenv(v):
                    vars = vars + "export %s=%s\n " % (v,os.getenv(v))
            vars = vars + "#" * 80 + "\n"

            script = script.replace("#SBATCH -v","###########SBATCH -v")
            #script = script.replace("\njupyterhub-singleuser","\n"+vars+"\njupyterhub-singleuser  --config=/home/kortass/JUPYTER/jupyterhub_config.py")
            script = script.replace("\njupyterhub-singleuser","\n"+vars+\
                                    "\nwhich jupyterhub-singleuser " +\
                                    "\njupyterhub-singleuser ")
            #script = script.replace('--notebook-dir="/home/kortass/JUPYTER/NOTEBOOKS"','')
            

        os.system("env > %s/jobs/%s_last_script_env" %  (kslhub_root,self.user.name))

        
        f = open("%s/jobs/%s_last_script_modified" % (kslhub_root,self.user.name),"w")
        f.write(script)
        f.close()


        self.log.info('Job content: \n' + '='*80 + '\n' + script + '\n' + '='*80 + '\n')
        
        # self.log.info('Spawner modified script:\n' + script)
        #SKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKkkk
        print("[AUTHENT] in submit_batch_script user: %s or |||%s|| ot type %s " %\
                  (self.user,pprint.pformat(self.user),type(self.user)))

        user = self.user
        user_str = ("%s" % user)
        if user_str.find('<User(')>-1:
            user = user_str[6:].split(' ')[0]



            
        out = yield run_command(cmd, input=script, env=self.get_env(), user=user)
        try:
            self.log.info('Job submitted. cmd: ' + cmd + ' output: ' + out)
            self.job_id = self.parse_job_id(out)
            f = open("%s/jobs/%s_%s.job" % (kslhub_root,self.user.name,self.job_id),"w")
            f.write(script)
            f.close()
            
        except:
            self.log.error('Job submission failed with exit code %s ' % out)
            self.job_id = ''
        return self.job_id

    # Override if your batch system needs something more elaborate to read the job status
    batch_query_cmd = Unicode('', config=True, \
        help="Command to run to read job status. Formatted using req_xyz traits as {xyz} "
             "and self.job_id as {job_id}."
        )

    @gen.coroutine
    def read_job_state(self):
        if self.job_id is None or len(self.job_id) == 0:
            # job not running
            self.job_status = ''
            return self.job_status
        subvars = self.get_req_subvars()
        subvars['job_id'] = self.job_id
        cmd = self.batch_query_cmd.format(**subvars)
        self.log.debug('Spawner querying job: ' + cmd)
        try:
            user = self.user
            user_str = ("%s" % user)
            if user_str.find('<User(')>-1:
                user = user_str[6:].split(' ')[0]
            out = yield run_command(cmd,user=user)
            self.job_status = out
        except Exception as e:
            self.log.error('Error querying job ' + self.job_id)
            self.job_status = ''
        finally:
            return self.job_status

    batch_cancel_cmd = Unicode('', config=True,
        help="Command to stop/cancel a previously submitted job. Formatted like batch_query_cmd."
        )

    @gen.coroutine
    def cancel_batch_job(self):
        subvars = self.get_req_subvars()
        subvars['job_id'] = self.job_id
        cmd = self.batch_cancel_cmd.format(**subvars)
        self.log.info('Cancelling job ' + self.job_id + ': ' + cmd)
        user = self.user
        user_str = ("%s" % user)
        if user_str.find('<User(')>-1:
            user = user_str[6:].split(' ')[0]
        yield run_command(cmd,user=user)

    def load_state(self, state):
        """load job_id from state"""
        super(BatchSpawnerBase, self).load_state(state)
        self.job_id = state.get('job_id', '')
        self.job_status = state.get('job_status', '')

    def get_state(self):
        """add job_id to state"""
        state = super(BatchSpawnerBase, self).get_state()
        if self.job_id:
            state['job_id'] = self.job_id
        if self.job_status:
            state['job_status'] = self.job_status
        return state

    def clear_state(self):
        """clear job_id state"""
        super(BatchSpawnerBase, self).clear_state()
        self.job_id = ""
        self.job_status = ''

    def make_preexec_fn(self, name):
        """make preexec fn to change uid (if running as root) before job submission"""
        return set_user_setuid(name)

    def state_ispending(self):
        "Return boolean indicating if job is still waiting to run, likely by parsing self.job_status"
        raise NotImplementedError("Subclass must provide implementation")

    def state_isrunning(self):
        "Return boolean indicating if job is running, likely by parsing self.job_status"
        raise NotImplementedError("Subclass must provide implementation")

    def state_gethost(self):
        "Return string, hostname or addr of running job, likely by parsing self.job_status"
        raise NotImplementedError("Subclass must provide implementation")

    @gen.coroutine
    def poll(self):
        """Poll the process"""
        if self.job_id is not None and len(self.job_id) > 0:
            yield self.read_job_state()
            if self.state_isrunning() or self.state_ispending():
                return None
            else:
                self.clear_state()
                return 1

        if not self.job_id:
            # no job id means it's not running
            self.clear_state()
            return 1

    startup_poll_interval = Float(5, config=True, \
        help="Polling interval (seconds) to check job state during startup"
        )

    @gen.coroutine
    def start(self):
        """Start the process"""
        if not self.user.server.port:
            self.user.server.port = random_port()
            self.db.commit()
        job = yield self.submit_batch_script()

        # We are called with a timeout, and if the timeout expires this function will
        # be interrupted at the next yield, and self.stop() will be called. 
        # So this function should not return unless successful, and if unsuccessful
        # should either raise and Exception or loop forever.
        assert len(self.job_id) > 0
        while True:
            yield self.poll()
            if self.state_isrunning():
                break
            else:
                if self.state_ispending():
                    self.log.debug('Job ' + self.job_id + ' still pending')
                else:
                    self.log.warn('Job ' + self.job_id + ' neither pending nor running.\n' +
                        self.job_status)
                assert self.state_ispending()
            yield gen.sleep(self.startup_poll_interval)

        self.user.server.ip = self.state_gethost()
        self.db.commit()
        self.log.info("Notebook server job {0} started at {1}:{2}".format(
                        self.job_id, self.user.server.ip, self.user.server.port)
            )

        return self.user.server.ip, self.user.server.port

    @gen.coroutine
    def stop(self, now=False):
        """Stop the singleuser server job.

        Returns immediately after sending job cancellation command if now=True, otherwise
        tries to confirm that job is no longer running."""

        self.log.info("Stopping server job " + self.job_id)
        yield self.cancel_batch_job()
        if now:
            return
        for i in range(10):
            yield self.poll()
            if not self.state_isrunning():
                return
            yield gen.sleep(1.0)
        if self.job_id:
            self.log.warn("Notebook server job {0} at {1}:{2} possibly failed to terminate".format(
                             self.job_id, self.user.server.ip, self.user.server.port)
                )


class BatchSpawnerRegexStates(BatchSpawnerBase):
    """Subclass of BatchSpawnerBase that uses config-supplied regular expressions
    to interact with batch submission system state. Provides implementations of
        state_ispending
        state_isrunning
        state_gethost

    In their place, the user should supply the following configuration:
        state_pending_re - regex that matches job_status if job is waiting to run
        state_running_re - regex that matches job_status if job is running
        state_exechost_re - regex with at least one capture group that extracts
                            execution host from job_status
        state_exechost_exp - if empty, notebook IP will be set to the contents of the
            first capture group. If this variable is set, the match object
            will be expanded using this string to obtain the notebook IP.
            See Python docs: re.match.expand
    """
    state_pending_re = Unicode('', config=True,
        help="Regex that matches job_status if job is waiting to run")
    state_running_re = Unicode('', config=True,
        help="Regex that matches job_status if job is running")
    state_exechost_re = Unicode('', config=True,
        help="Regex with at least one capture group that extracts "
             "the execution host from job_status output")
    state_exechost_exp = Unicode('', config=True,
        help="""If empty, notebook IP will be set to the contents of the first capture group.

        If this variable is set, the match object will be expanded using this string
        to obtain the notebook IP.
        See Python docs: re.match.expand""")

    def state_ispending(self):
        assert self.state_pending_re
        if self.job_status and re.search(self.state_pending_re, self.job_status):
            return True
        else: return False

    def state_isrunning(self):
        assert self.state_running_re
        self.log.info("is_running job_status=/%s/ is_running =/%s/" % \
                      (self.job_status,re.search(self.state_running_re, self.job_status)))

        if self.job_status and re.search(self.state_running_re, self.job_status):
            return True
        else: return False

    def state_gethost(self):
        assert self.state_exechost_re
        match = re.search(self.state_exechost_re, self.job_status)
        self.log.info("is_host job_status=/%s/ host =/%s/" % (self.job_status,match.groups()[0]))

        if not match:
            self.log.error("Spawner unable to match host addr in job status: " + self.job_status)
            return
        if not self.state_exechost_exp:
            return match.groups()[0].replace('sam-1','localhost')
        else:
            return match.expand(self.state_exechost_exp)

class UserEnvMixin:
    """Mixin class that computes values for USER, SHELL and HOME in the environment passed to
    the job submission subprocess in case the batch system needs these for the batch script."""

    def user_env(self, env):
        """get user environment"""
        env['USER'] = self.user.name
        home = pwd.getpwnam(self.user.name).pw_dir
        shell = pwd.getpwnam(self.user.name).pw_shell
        if home:
            env['HOME'] = home
        if shell:
            env['SHELL'] = shell
        return env

    def get_env(self):
        """Add user environment variables"""
        env = super().get_env()
        env = self.user_env(env)
        return env

class SlurmSpawner(UserEnvMixin,BatchSpawnerRegexStates):
    """A Spawner that just uses Popen to start local processes."""

    # all these req_foo traits will be available as substvars for templated strings
    req_partition = Unicode('', config=True, \
        help="Partition name to submit job to resource manager"
        )

    req_qos = Unicode('', config=True, \
        help="QoS name to submit job to resource manager"
        )

    batch_script = Unicode("""#!/bin/bash
#SBATCH --partition={partition}
#SBATCH --time={runtime}
#SBATCH --output={homedir}/jupyterhub_slurmspawner_%j.log
#SBATCH --job-name=spawner-jupyterhub
#SBATCH --workdir={homedir}
#SBATCH --mem={memory}
#SBATCH --export={keepvars}
#SBATCH --uid={username}
#SBATCH --get-user-env=L
#SBATCH {options}

which jupyterhub-singleuser
{cmd}
""",
        config=True)

    if slurm_machine.machine == "locxalhost":
        # outputs line like "Submitted batch job 209"
        batch_submit_cmd = Unicode('/home/kortass/opt/slurm/bin/sbatch', config=True)    
        # outputs status and exec node like "RUNNING hostname"
        batch_query_cmd = Unicode('/home/kortass/opt/slurm/bin/squeue -h -j {job_id} -o "%T %B"', config=True) #
        batch_cancel_cmd = Unicode('/home/kortass/opt/slurm/bin/scancel {job_id}', config=True)

    else:
        batch_submit_cmd = Unicode('sbatch', config=True)
        batch_query_cmd = Unicode('squeue -h -j {job_id} -o "%T %B"', config=True) 
        batch_cancel_cmd = Unicode('scancel {job_id}', config=True)
    

    # use long-form states: PENDING,  CONFIGURING = pending
    #  RUNNING,  COMPLETING = running
    state_pending_re = Unicode(r'^(?:PENDING|CONFIGURING)', config=True)
    state_running_re = Unicode(r'^(?:RUNNING|COMPLETING)', config=True)
    state_exechost_re = Unicode(r'\s+((?:[\w_-]+\.?)+)$', config=True)

    
    def parse_job_id(self, output):
        # make sure jobid is really a number
        try:
            id = output.split(' ')[-1]
            int(id)
        except Exception as e:
            self.log.error("SlurmSpawner unable to parse job ID from text: " + output)
            raise e
        return id

class MultiSlurmSpawner(SlurmSpawner):
    '''When slurm has been compiled with --enable-multiple-slurmd, the
       administrator sets the name of the slurmd instance via the slurmd -N
       option. This node name is usually different from the hostname and may
       not be resolvable by JupyterHub. Here we enable the administrator to
       map the node names onto the real hostnames via a traitlet.'''
    daemon_resolver = Dict({}, config=True, help="Map node names to hostnames")

    def state_gethost(self):
        host = SlurmSpawner.state_gethost(self)
        return self.daemon_resolver.get(host, host)

# vim: set ai expandtab softtabstop=4:



c.JupyterHub.spawner_class = 'wrapspawner.ProfilesSpawner'
c.Spawner.http_timeout = 120
#------------------------------------------------------------------------------
# ProfilesSpawner configuration
#------------------------------------------------------------------------------
# List of profiles to offer for selection. Signature is:
#   List(Tuple( Unicode, Unicode, Type(Spawner), Dict ))
# corresponding to profile display name, unique key, Spawner class,
# dictionary of spawner config options.
# 
# The first three values will be exposed in the input_template as {display},
# {key}, and {type}
#
c.ProfilesSpawner.profiles = [
       ( "Host process", 'local', 'jupyterhub.spawner.LocalProcessSpawner', {'ip':'0.0.0.0'} ),
 ]

 



"""sqlalchemy ORM tools for the state of the constellation of processes"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from datetime import datetime, timedelta
import enum
import json

import alembic.config
import alembic.command
from alembic.script import ScriptDirectory
from tornado.log import app_log

from sqlalchemy.types import TypeDecorator, TEXT, LargeBinary
from sqlalchemy import (
    create_engine, event, exc, inspect, or_, select,
    Column, Integer, ForeignKey, Unicode, Boolean,
    DateTime, Enum, Table,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.interfaces import PoolListener
from sqlalchemy.orm import (
    Session,
    interfaces, object_session, relationship, sessionmaker,
)

from sqlalchemy.pool import StaticPool
from sqlalchemy.sql.expression import bindparam

from .utils import (
    random_port,
    new_token, hash_token, compare_token,
)

# top-level variable for easier mocking in tests
utcnow = datetime.utcnow


class JSONDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::

        JSONEncodedDict(255)

    """

    impl = TEXT

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)

        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


Base = declarative_base()
Base.log = app_log


class Server(Base):
    """The basic state of a server

    connection and cookie info
    """
    __tablename__ = 'servers'
    id = Column(Integer, primary_key=True)

    proto = Column(Unicode(15), default='http')
    ip = Column(Unicode(255), default='')  # could also be a DNS name
    port = Column(Integer, default=random_port)
    base_url = Column(Unicode(255), default='/')
    cookie_name = Column(Unicode(255), default='cookie')

    def __repr__(self):
        return "<Server(%s:%s)>" % (self.ip, self.port)


# user:group many:many mapping table
user_group_map = Table('user_group_map', Base.metadata,
    Column('user_id', ForeignKey('users.id', ondelete='CASCADE'), primary_key=True),
    Column('group_id', ForeignKey('groups.id', ondelete='CASCADE'), primary_key=True),
)


class Group(Base):
    """User Groups"""
    __tablename__ = 'groups'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(Unicode(255), unique=True)
    users = relationship('User', secondary='user_group_map', backref='groups')

    def __repr__(self):
        return "<%s %s (%i users)>" % (
            self.__class__.__name__, self.name, len(self.users)
        )

    @classmethod
    def find(cls, db, name):
        """Find a group by name.
        Returns None if not found.
        """
        return db.query(cls).filter(cls.name == name).first()


class User(Base):
    """The User table

    Each user can have one or more single user notebook servers.

    Each single user notebook server will have a unique token for authorization.
    Therefore, a user with multiple notebook servers will have multiple tokens.

    API tokens grant access to the Hub's REST API.
    These are used by single-user servers to authenticate requests,
    and external services to manipulate the Hub.

    Cookies are set with a single ID.
    Resetting the Cookie ID invalidates all cookies, forcing user to login again.

    A `state` column contains a JSON dict,
    used for restoring state of a Spawner.


    `servers` is a list that contains a reference for each of the user's single user notebook servers.
    The method `server` returns the first entry in the user's `servers` list.
    """
    __tablename__ = 'users'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(Unicode(255), unique=True)

    _orm_spawners = relationship(
        "Spawner",
        backref="user",
        cascade="all, delete-orphan",
    )
    @property
    def orm_spawners(self):
        return {s.name: s for s in self._orm_spawners}

    admin = Column(Boolean, default=False)
    created = Column(DateTime, default=datetime.utcnow)
    last_activity = Column(DateTime, nullable=True)

    api_tokens = relationship(
        "APIToken",
        backref="user",
        cascade="all, delete-orphan",
    )
    oauth_tokens = relationship(
        "OAuthAccessToken",
        backref="user",
        cascade="all, delete-orphan",
    )
    oauth_codes = relationship(
        "OAuthCode",
        backref="user",
        cascade="all, delete-orphan",
    )
    cookie_id = Column(Unicode(255), default=new_token, nullable=False, unique=True)
    # User.state is actually Spawner state
    # We will need to figure something else out if/when we have multiple spawners per user
    state = Column(JSONDict)
    # Authenticators can store their state here:
    # Encryption is handled elsewhere
    encrypted_auth_state = Column(LargeBinary)

    def __repr__(self):
        return "<{cls}({name} {running}/{total} running)>".format(
            cls=self.__class__.__name__,
            name=self.name,
            total=len(self._orm_spawners),
            running=sum(bool(s.server) for s in self._orm_spawners),
        )

    def new_api_token(self, token=None, **kwargs):
        """Create a new API token

        If `token` is given, load that token.
        """
        #SKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKK
        #return "api_token"
        return APIToken.new(token=token, user=self, **kwargs)

    @classmethod
    def find(cls, db, name):
        """Find a user by name.
        Returns None if not found.
        """
        return db.query(cls).filter(cls.name == name).first()

class Spawner(Base):
    """"State about a Spawner"""
    __tablename__ = 'spawners'

    id = Column(Integer, primary_key=True, autoincrement=True)
    user_id = Column(Integer, ForeignKey('users.id', ondelete='CASCADE'))

    server_id = Column(Integer, ForeignKey('servers.id', ondelete='SET NULL'))
    server = relationship(Server, cascade="all")

    state = Column(JSONDict)
    name = Column(Unicode(255))

    started = Column(DateTime)
    last_activity = Column(DateTime, nullable=True)

    # properties on the spawner wrapper
    # some APIs get these low-level objects
    # when the spawner isn't running,
    # for which these should all be False
    active = running = ready = False
    pending = None
    @property
    def orm_spawner(self):
        return self

class Service(Base):
    """A service run with JupyterHub

    A service is similar to a User without a Spawner.
    A service can have API tokens for accessing the Hub's API

    It has:
    - name
    - admin
    - api tokens
    - server (if proxied http endpoint)

    In addition to what it has in common with users, a Service has extra info:

    - pid: the process id (if managed)

    """
    __tablename__ = 'services'
    id = Column(Integer, primary_key=True, autoincrement=True)

    # common user interface:
    name = Column(Unicode(255), unique=True)
    admin = Column(Boolean, default=False)

    api_tokens = relationship(
        "APIToken",
        backref="service",
        cascade="all, delete-orphan",
    )

    # service-specific interface
    _server_id = Column(Integer, ForeignKey('servers.id', ondelete='SET NULL'))
    server = relationship(Server, cascade='all')
    pid = Column(Integer)

    def new_api_token(self, token=None, **kwargs):
        """Create a new API token
        If `token` is given, load that token.
        """
        #SKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKK
        #return "api_token"
        return APIToken.new(token=token, service=self, **kwargs)

    @classmethod
    def find(cls, db, name):
        """Find a service by name.

        Returns None if not found.
        """
        return db.query(cls).filter(cls.name == name).first()


class Hashed(object):
    """Mixin for tables with hashed tokens"""
    prefix_length = 4
    algorithm = "sha512"
    rounds = 16384
    salt_bytes = 8
    min_length = 8

    # values to use for internally generated tokens,
    # which have good entropy as UUIDs
    generated = True
    generated_salt_bytes = 8
    generated_rounds = 1

    @property
    def token(self):
        raise AttributeError("token is write-only")

    @token.setter
    def token(self, token):
        """Store the hashed value and prefix for a token"""
        self.prefix = token[:self.prefix_length]
        if self.generated:
            # Generated tokens are UUIDs, which have sufficient entropy on their own
            # and don't need salt & hash rounds.
            # ref: https://security.stackexchange.com/a/151262/155114
            rounds = self.generated_rounds
            salt_bytes = self.generated_salt_bytes
        else:
            rounds = self.rounds
            salt_bytes = self.salt_bytes
        self.hashed = hash_token(token, rounds=rounds, salt=salt_bytes, algorithm=self.algorithm)

    def match(self, token):
        """Is this my token?"""
        return compare_token(self.hashed, token)

    @classmethod
    def check_token(cls, db, token):
        """Check if a token is acceptable"""
        if len(token) < cls.min_length:
            raise ValueError("Tokens must be at least %i characters, got %r" % (
                cls.min_length, token)
            )
        found = cls.find(db, token)
        if found:
            raise ValueError("Collision on token: %s..." % token[:cls.prefix_length])

    @classmethod
    def find_prefix(cls, db, token):
        """Start the query for matching token.

        Returns an SQLAlchemy query already filtered by prefix-matches.
        """
        prefix = token[:cls.prefix_length]
        # since we can't filter on hashed values, filter on prefix
        # so we aren't comparing with all tokens
        return db.query(cls).filter(bindparam('prefix', prefix).startswith(cls.prefix))

    @classmethod
    def find(cls, db, token):
        """Find a token object by value.

        Returns None if not found.

        `kind='user'` only returns API tokens for users
        `kind='service'` only returns API tokens for services
        """
        prefix_match = cls.find_prefix(db, token)
        for orm_token in prefix_match:
            if orm_token.match(token):
                return orm_token


class APIToken(Hashed, Base):
    """An API token"""
    __tablename__ = 'api_tokens'

    user_id = Column(Integer, ForeignKey('users.id', ondelete="CASCADE"), nullable=True)
    service_id = Column(Integer, ForeignKey('services.id', ondelete="CASCADE"), nullable=True)

    id = Column(Integer, primary_key=True)
    hashed = Column(Unicode(255), unique=True)
    prefix = Column(Unicode(16), index=True)

    @property
    def api_id(self):
        return 'a%i' % self.id

    # token metadata for bookkeeping
    created = Column(DateTime, default=datetime.utcnow)
    expires_at = Column(DateTime, default=None, nullable=True)
    last_activity = Column(DateTime)
    note = Column(Unicode(1023))

    def __repr__(self):
        if self.user is not None:
            kind = 'user'
            name = self.user.name
        elif self.service is not None:
            kind = 'service'
            name = self.service.name
        else:
            # this shouldn't happen
            kind = 'owner'
            name = 'unknown'
        return "<{cls}('{pre}...', {kind}='{name}')>".format(
            cls=self.__class__.__name__,
            pre=self.prefix,
            kind=kind,
            name=name,
        )

    @classmethod
    def purge_expired(cls, db):
        """Purge expired API Tokens from the database"""
        now = utcnow()
        deleted = False
        for token in (
            db.query(cls)
            .filter(cls.expires_at != None)
            .filter(cls.expires_at < now)
        ):
            app_log.debug("Purging expired %s", token)
            deleted = True
            db.delete(token)
        if deleted:
            db.commit()

    @classmethod
    def find(cls, db, token, *, kind=None):
        """Find a token object by value.

        Returns None if not found.

        `kind='user'` only returns API tokens for users
        `kind='service'` only returns API tokens for services
        """
        prefix_match = cls.find_prefix(db, token)
        prefix_match = prefix_match.filter(
            or_(cls.expires_at == None, cls.expires_at >= utcnow())
        )
        if kind == 'user':
            prefix_match = prefix_match.filter(cls.user_id != None)
        elif kind == 'service':
            prefix_match = prefix_match.filter(cls.service_id != None)
        elif kind is not None:
            raise ValueError("kind must be 'user', 'service', or None, not %r" % kind)
        for orm_token in prefix_match:
            if orm_token.match(token):
                return orm_token

    @classmethod
    def new(cls, token=None, user=None, service=None, note='', generated=True,
            expires_in=None):
        """Generate a new API token for a user or service"""
        assert user or service
        assert not (user and service)
        db = inspect(user or service).session
        if token is None:
            token = new_token()
            # Don't need hash + salt rounds on generated tokens,
            # which already have good entropy
            generated = True
        else:
            cls.check_token(db, token)
        # two stages to ensure orm_token.generated has been set
        # before token setter is called
        orm_token = cls(generated=generated, note=note or '')
        orm_token.token = token
        if user:
            assert user.id is not None
            orm_token.user = user
        else:
            assert service.id is not None
            orm_token.service = service
        if expires_in is not None:
            orm_token.expires_at = utcnow() + timedelta(seconds=expires_in)
        db.add(orm_token)
        db.commit()
        return token


#------------------------------------
# OAuth tables
#------------------------------------


class GrantType(enum.Enum):
    # we only use authorization_code for now
    authorization_code = 'authorization_code'
    implicit = 'implicit'
    password = 'password'
    client_credentials = 'client_credentials'
    refresh_token = 'refresh_token'


class OAuthAccessToken(Hashed, Base):
    __tablename__ = 'oauth_access_tokens'
    id = Column(Integer, primary_key=True, autoincrement=True)

    @property
    def api_id(self):
        return 'o%i' % self.id

    client_id = Column(Unicode(255), ForeignKey('oauth_clients.identifier', ondelete='CASCADE'))
    grant_type = Column(Enum(GrantType), nullable=False)
    expires_at = Column(Integer)
    refresh_token = Column(Unicode(255))
    # TODO: drop refresh_expires_at. Refresh tokens shouldn't expire
    refresh_expires_at = Column(Integer)
    user_id = Column(Integer, ForeignKey('users.id', ondelete='CASCADE'))
    service = None # for API-equivalence with APIToken

    # the browser session id associated with a given token
    session_id = Column(Unicode(255))

    # from Hashed
    hashed = Column(Unicode(255), unique=True)
    prefix = Column(Unicode(16), index=True)

    created = Column(DateTime, default=datetime.utcnow)
    last_activity = Column(DateTime, nullable=True)

    def __repr__(self):
        return "<{cls}('{prefix}...', client_id={client_id!r}, user={user!r}>".format(
            cls=self.__class__.__name__,
            client_id=self.client_id,
            user=self.user and self.user.name,
            prefix=self.prefix,
        )

    @classmethod
    def find(cls, db, token):
        orm_token = super().find(db, token)
        if orm_token and not orm_token.client_id:
            app_log.warning(
                "Deleting stale oauth token for %s with no client",
                orm_token.user and orm_token.user.name,
            )
            db.delete(orm_token)
            db.commit()
            return
        return orm_token


class OAuthCode(Base):
    __tablename__ = 'oauth_codes'
    id = Column(Integer, primary_key=True, autoincrement=True)
    client_id = Column(Unicode(255), ForeignKey('oauth_clients.identifier', ondelete='CASCADE'))
    code = Column(Unicode(36))
    expires_at = Column(Integer)
    redirect_uri = Column(Unicode(1023))
    session_id = Column(Unicode(255))
    # state = Column(Unicode(1023))
    user_id = Column(Integer, ForeignKey('users.id', ondelete='CASCADE'))


class OAuthClient(Base):
    __tablename__ = 'oauth_clients'
    id = Column(Integer, primary_key=True, autoincrement=True)
    identifier = Column(Unicode(255), unique=True)
    description = Column(Unicode(1023))
    secret = Column(Unicode(255))
    redirect_uri = Column(Unicode(1023))

    @property
    def client_id(self):
        return self.identifier

    access_tokens = relationship(
        OAuthAccessToken,
        backref='client',
        cascade='all, delete-orphan',
    )
    codes = relationship(
        OAuthCode,
        backref='client',
        cascade='all, delete-orphan',
    )

# General database utilities

class DatabaseSchemaMismatch(Exception):
    """Exception raised when the database schema version does not match

    the current version of JupyterHub.
    """


class ForeignKeysListener(PoolListener):
    """Enable foreign keys on sqlite"""
    def connect(self, dbapi_con, con_record):
        dbapi_con.execute('pragma foreign_keys=ON')


def _expire_relationship(target, relationship_prop):
    """Expire relationship backrefs

    used when an object with relationships is deleted
    """

    session = object_session(target)
    # get peer objects to be expired
    peers = getattr(target, relationship_prop.key)
    if peers is None:
        # no peer to clear
        return
    # many-to-many and one-to-many have a list of peers
    # many-to-one has only one
    if relationship_prop.direction is interfaces.MANYTOONE:
        peers = [peers]
    for obj in peers:
        if inspect(obj).persistent:
            session.expire(obj, [relationship_prop.back_populates])


@event.listens_for(Session, "persistent_to_deleted")
def _notify_deleted_relationships(session, obj):
    """Expire relationships when an object becomes deleted

    Needed to keep relationships up to date.
    """
    mapper = inspect(obj).mapper
    for prop in mapper.relationships:
        if prop.back_populates:
            _expire_relationship(obj, prop)


def register_ping_connection(engine):
    """Check connections before using them.

    Avoids database errors when using stale connections.

    From SQLAlchemy docs on pessimistic disconnect handling:

    https://docs.sqlalchemy.org/en/rel_1_1/core/pooling.html#disconnect-handling-pessimistic
    """
    @event.listens_for(engine, "engine_connect")
    def ping_connection(connection, branch):
        if branch:
            # "branch" refers to a sub-connection of a connection,
            # we don't want to bother pinging on these.
            return

        # turn off "close with result".  This flag is only used with
        # "connectionless" execution, otherwise will be False in any case
        save_should_close_with_result = connection.should_close_with_result
        connection.should_close_with_result = False

        try:
            # run a SELECT 1.   use a core select() so that
            # the SELECT of a scalar value without a table is
            # appropriately formatted for the backend
            connection.scalar(select([1]))
        except exc.DBAPIError as err:
            # catch SQLAlchemy's DBAPIError, which is a wrapper
            # for the DBAPI's exception.  It includes a .connection_invalidated
            # attribute which specifies if this connection is a "disconnect"
            # condition, which is based on inspection of the original exception
            # by the dialect in use.
            if err.connection_invalidated:
                app_log.error("Database connection error, attempting to reconnect: %s", err)
                # run the same SELECT again - the connection will re-validate
                # itself and establish a new connection.  The disconnect detection
                # here also causes the whole connection pool to be invalidated
                # so that all stale connections are discarded.
                connection.scalar(select([1]))
            else:
                raise
        finally:
            # restore "close with result"
            connection.should_close_with_result = save_should_close_with_result


def check_db_revision(engine):
    """Check the JupyterHub database revision

    After calling this function, an alembic tag is guaranteed to be stored in the db.

    - Checks the alembic tag and raises a ValueError if it's not the current revision
    - If no tag is stored (Bug in Hub prior to 0.8),
      guess revision based on db contents and tag the revision.
    - Empty databases are tagged with the current revision
    """
    # Check database schema version
    current_table_names = set(engine.table_names())
    my_table_names = set(Base.metadata.tables.keys())

    from .dbutil import _temp_alembic_ini

    with _temp_alembic_ini(engine.url) as ini:
        cfg = alembic.config.Config(ini)
        scripts = ScriptDirectory.from_config(cfg)
        head = scripts.get_heads()[0]
        base = scripts.get_base()

        if not my_table_names.intersection(current_table_names):
            # no tables have been created, stamp with current revision
            app_log.debug("Stamping empty database with alembic revision %s", head)
            alembic.command.stamp(cfg, head)
            return

        if 'alembic_version' not in current_table_names:
            # Has not been tagged or upgraded before.
            # we didn't start tagging revisions correctly except during `upgrade-db`
            # until 0.8
            # This should only occur for databases created prior to JupyterHub 0.8
            msg_t = "Database schema version not found, guessing that JupyterHub %s created this database."
            if 'spawners' in current_table_names:
                # 0.8
                app_log.warning(msg_t, '0.8.dev')
                rev = head
            elif 'services' in current_table_names:
                # services is present, tag for 0.7
                app_log.warning(msg_t, '0.7.x')
                rev = 'af4cbdb2d13c'
            else:
                # it's old, mark as first revision
                app_log.warning(msg_t, '0.6 or earlier')
                rev = base
            app_log.debug("Stamping database schema version %s", rev)
            alembic.command.stamp(cfg, rev)

    # check database schema version
    # it should always be defined at this point
    alembic_revision = engine.execute('SELECT version_num FROM alembic_version').first()[0]
    if alembic_revision == head:
        app_log.debug("database schema version found: %s", alembic_revision)
        pass
    else:
        raise DatabaseSchemaMismatch("Found database schema version {found} != {head}. "
        "Backup your database and run `jupyterhub upgrade-db`"
        " to upgrade to the latest schema.".format(
            found=alembic_revision,
            head=head,
        ))


def mysql_large_prefix_check(engine):
    """Check mysql has innodb_large_prefix set"""
    if not str(engine.url).startswith('mysql'):
        return False
    variables =  dict(engine.execute(
        'show variables where variable_name like '
        '"innodb_large_prefix" or '
        'variable_name like "innodb_file_format";').fetchall())
    if (variables['innodb_file_format'] == 'Barracuda' and
        variables['innodb_large_prefix'] == 'ON'):
        return True
    else:
        return False


def add_row_format(base):
    for t in  base.metadata.tables.values():
        t.dialect_kwargs['mysql_ROW_FORMAT'] = 'DYNAMIC'


def new_session_factory(url="sqlite:///:memory:",
                        reset=False,
                        expire_on_commit=False,
                        **kwargs):
    """Create a new session at url"""
    if url.startswith('sqlite'):
        kwargs.setdefault('connect_args', {'check_same_thread': False})
        listeners = kwargs.setdefault('listeners', [])
        listeners.append(ForeignKeysListener())

    elif url.startswith('mysql'):
        kwargs.setdefault('pool_recycle', 60)

    if url.endswith(':memory:'):
        # If we're using an in-memory database, ensure that only one connection
        # is ever created.
        kwargs.setdefault('poolclass', StaticPool)

    engine = create_engine(url, **kwargs)
    # enable pessimistic disconnect handling
    register_ping_connection(engine)

    if reset:
        Base.metadata.drop_all(engine)

    if mysql_large_prefix_check(engine):  # if mysql is allows large indexes
        add_row_format(Base)              # set format on the tables
    # check the db revision (will raise, pointing to `upgrade-db` if version doesn't match)
    check_db_revision(engine)

    Base.metadata.create_all(engine)

    # We set expire_on_commit=False, since we don't actually need
    # SQLAlchemy to expire objects after committing - we don't expect
    # concurrent runs of the hub talking to the same db. Turning
    # this off gives us a major performance boost
    session_factory = sessionmaker(bind=engine,
                                   expire_on_commit=expire_on_commit,
    )
    return session_factory



import paramiko
import os,re, sys, traceback
import pprint

SSH_DEBUG=False
HOME = os.getenv("HOME")

class session_manager(object):
    __shared_state = {}
    
    def __init__(self):
        self.__dict__ = self.__shared_state
        if len(self.__dict__)==0:
            self.connections = {}
            self.passwords = {}
            self.machine = None

    def initialize_machine(self,machine):
        self.machine = machine

        # on lit les user par defaut dans .ssh/config
        # la clef de connection elle par contre est en dur : ~/.ssh/id_rsa-hpcportal
        ssh_config = paramiko.SSHConfig()
        try:
            ssh_config.parse(open(os.path.abspath(HOME)+'/.ssh/config'))
            o = ssh_config.lookup(self.machine)
            if SSH_DEBUG:
                print (self.machine,o)
            self.actual_machine = o['hostname']
            #self.actual_default_user = o["user"]
            self.actual_default_port = 22
            try:
                if "port" in o.keys():
                    self.actual_default_port = int(o["port"])
            except:
                print('ZZZZ problem to read port ',o)
    
    
            if SSH_DEBUG:
                print("actual_machine: %s, actual_default_port: %s " % \
                      (self.actual_machine, self.actual_default_port))
            # bug bizzare de lecture du user
            #self.actual_default_user = self.actual_default_user[:-1]
            # avec un blanc sur certaines machines
            # self.actual_ssh = o['identityfile']
            
        except Exception as e:
            print("ssh_authenticator.py : problem to read ssh config for machine %s" %\
                  self.machine, " read from : %s" % \
                  os.path.abspath(HOME)+'/.ssh/config')
            try:
                print("value read from .ssh/config : ",o)
            except:
                print("no value read from .ssh/config....")
            print(e)
            traceback.print_exc(file=sys.stdout)
            print("setting same machine name and port to default ssh port")
            #sys.exit(1)
            self.actual_machine = machine
            self.actual_default_port = 22



    def open(self,machine,login,password):
        if not(self.machine):
            self.initialize_machine(machine)
            
        user_connection_id = "%s_%s" % (self.machine,login)
        
        if SSH_DEBUG:
            print("[AUTHENT] opens new session for  %s/%s" % \
                  (self.machine,login))

        # initialisation du client
        connection = paramiko.SSHClient()
        connection.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # login utilisateur
        try:
            connection.connect(self.actual_machine, username=login, password=password,
                               port=self.actual_default_port)
            if SSH_DEBUG:
                print("[AUTHENT] connection succeeded!!!")
        except Exception as e:
            if SSH_DEBUG:
                print("[AUTHENT] connection failed!!!")
            if user_connection_id in self.connections.keys():
                del self.connections[user_connection_id] 
            if user_connection_id in self.passwords.keys():
                del self.passwords[user_connection_id]
            print(e)
            traceback.print_exc(file=sys.stdout)

            return False
            raise ValueError("bad connection")


        if SSH_DEBUG:
            print("[AUTHENT] [%s] opens new connection for user %s : " % \
                (self.machine,login) ,connection)
        # test de la connection
        stdin, stdout, stderr = connection.exec_command("cd; pwd")
        res = stdout.readlines()
        home = res
        if SSH_DEBUG:
            print("[AUTHENT] stdout test : ",res)
            print("[AUTHENT] stderr test ; ",stderr.readlines())
                
        self.connections[user_connection_id] = connection
        self.passwords[user_connection_id] = password

        if SSH_DEBUG:
            print("[AUTHENT] so far connection registered are %s",pprint.pformat(self.connections))

        return connection
        

    def execute(self,login,cmd):
        connection = self.get_session(login)
        if not(connection):
            print("[AUTHENT] session expired for %s/%s " %\
                  (self.machine,login))
            return False
        # test de la connection
        stdin, stdout, stderr = connection.exec_command(cmd)
        res = stdout.readlines()
        home = res
        return home

    def get_session(self,login):
        user_connection_id = "%s_%s" % (self.machine,login)
        if SSH_DEBUG:
            print("[AUTHENT] so far connection registered are %s",pprint.pformat(self.connections))
        if user_connection_id in self.connections.keys():
            if SSH_DEBUG:
                print("[AUTHENT] reutilisation de la session pour ",user_connection_id)
            return self.connections[user_connection_id]
        else:
            print("[AUTHENT] NO session available for ",user_connection_id)
            return False
        
    def close(self,login):
        user_connection_id = "%s_%s" % (self.machine,login)
        if user_connection_id in self.connections.keys():
            if SSH_DEBUG:
                print("[AUTHENT] closing ssh session for ",user_connection_id)
            # closing session
            try:
                if self.machine == "vishnu":
                    session_id = self.connections[user_connection_id]
                    VISHNU.close(session_id)
                else:
                    self.connections[user_connection_id].close()
            except:
                if SSH_DEBUG:
                    print("[AUTHENT] session %s already closed!!!!!" % user_connection_id)
            # cleaning
            del self.connections[user_connection_id]
            if user_connection_id in self.passwords.keys():
                 del self.passwords[user_connection_id]
            return True
        else:
            print("[AUTHENT] asking to close an unexisting  session for  ",user_connection_id)
            return False
        

    def exists(self,login):
        user_connection_id = "%s_%s" % (self.machine,login)
        if SSH_DEBUG:
            print ("[AUTHENT] Checking if sessions ssh exists  pour ",user_connection_id\
                ,user_connection_id in self.connections.keys())
        return (user_connection_id in self.connections.keys())
    





if __name__ == "__main__":

    user_name = "ek"
    slurm_machine = session_manager()
    connection = slurm_machine.open(user_name,"ksl123")

    if not(connection):
        print("[AUTHENT] session failed for %s/%s " %\
              (slurm_machine.machine,user_name))
        sys.exit(1)

    print("".join(slurm_machine.execute(user_name,"ps -edf")))

    connection = slurm_machine.get_session(user_name)
    if not(connection):
        print("[AUTHENT] session expired for %s/%s " %\
              (slurm_machine.machine,user_name))
        sys.exit(1)

    input = """#!/bin/sh
#SBATCH -N 1
#SBATCH -J job
#SBATCH -o job.out
#SBATCH -o job.err
#SBATCH -t 2
echo hello
sleep 10
"""
    cmd = "sbatch"
    stdin, stdout, stderr = connection.exec_command(cmd)
    stdin.write(input)
    stdin.channel.shutdown_write()
    res = stdout.readlines()
    print("[AUTHENT] res for cmd=|||%s||| = %s " %    (cmd,res))



import os
import sys

c.Authenticator.hub_greeting_message = "Welcome to KSL Hub @ docker!"
c.Authenticator.hub_name = "Kslhub@docker"
c.Authenticator.job_template_dir = "./job_templates"
current_host = 'kslhub'
#c.JupyterHub.hub_bind_url = 'http://%s:9001' % current_host

c.JupyterHub.authenticator_class = 'kslhub.otp_authenticator.SshUserAuthenticator'
c.Authenticator.otp_required = False 
c.Authenticator.host = 'localhost'

c.JupyterHub.template_paths = [ "%s/share/kslhub/templates" % sys.base_prefix ]

c.Spawner.cmd = ['jupyter-labhub']

runtime_dir = os.path.join('/tmp')


c.JupyterHub.spawner_class = 'kslhub.wrapspawner.ProfilesSpawner'
c.Spawner.http_timeout = 120
c.BatchSpawnerBase.req_host = 'localhost'

# c.ProfilesSpawner.profiles = [
#    ('Slurm - 1 Node - debug queue - 1 minutes', 'debug01', 'batchspawner.SlurmSpawner',
#       dict(req_nprocs='1', req_partition='debug', req_runtime='0:01:00')),
#    ('Slurm - 1 Node - debug queue - 2 minutes', 'debug02', 'batchspawner.SlurmSpawner',
#       dict(req_nprocs='1', req_partition='debug', req_runtime='0:02:00')),
#    ('Slurm - 1 Node - debug queue - 5 minutes', 'debug05', 'batchspawner.SlurmSpawner',
#       dict(req_nprocs='1', req_partition='debug', req_runtime='0:05:00')),
#    ('Slurm - 1 Node - debug queue - 30 minutes', 'debug30', 'batchspawner.SlurmSpawner',
#       dict(req_nprocs='1', req_partition='debug', req_runtime='0:30:00')),
#    ( "Local server", 'local', 'jupyterhub.spawner.LocalProcessSpawner', {'ip':'0.0.0.0'} ),
#    ]




# Copyright (c) Regents of the University of Minnesota
# Distributed under the terms of the Modified BSD License.

"""WrapSpawner classes

WrapSpawner provides a mechanism to wrap the interface of a Spawner such that
the Spawner class to use for single-user servers can be chosen dynamically.
The child Spawner is created and started using the same logic as in User.spawn(),
but not until start() or load_state() is called. Thus, subclasses may modify the
class or properties of the child Spawner at any earlier point (e.g. from
Authenticator pre_spawn hooks or options form processing).

Note that there is no straightforward way to simply save the child_class Type
value in the ORM, so a subclass that provides a way to change the child_class
must also arrange to reproduce that change when state is reloaded.

Provided as an initial application is ProfilesSpawner, which accepts a list
of provided Spawner configurations, and generates an options form from that list.
Using this mechanism, the administrator can provide users with a pre-approved
selection of Spawner configurations.
"""

import os
import json
import re
import urllib.request

import pprint

from tornado import gen, concurrent

from jupyterhub.spawner import LocalProcessSpawner, Spawner
import kslhub.batchspawner


from traitlets import (
    Instance, Type, Tuple, List, Dict, Integer, Unicode, Float, Any
)

# Only needed for DockerProfilesSpawner
try:
    import docker
except ImportError:
    pass

# Utility to create dummy Futures to return values through yields
def _yield_val(x=None):
    f = concurrent.Future()
    f.set_result(x)
    return f

class WrapSpawner(Spawner):

    # Grab this from constructor args in case some Spawner ever wants it
    config = Any()

    child_class = Type(LocalProcessSpawner, Spawner,
        config=True,
        help="""The class to wrap for spawning single-user servers.
                Should be a subclass of Spawner.
                """
        )

    child_config = Dict(default_value={},
        config=True,
        help="Dictionary of config values to apply to wrapped spawner class."
        )

    child_state = Dict(default_value={})

    child_spawner = Instance(Spawner, allow_none=True)

    def construct_child(self):
        if self.child_spawner is None:
            self.child_spawner = self.child_class(
                user = self.user,
                db   = self.db,
                hub  = self.hub,
                authenticator = self.authenticator,
                oauth_client_id = self.oauth_client_id,
                server = self._server,
                config = self.config,
                **self.child_config
                )
            # initial state will always be wrong since it will see *our* state
            self.child_spawner.clear_state()
            if self.child_state:
                self.child_spawner.load_state(self.child_state)
            self.child_spawner.api_token = self.api_token
        return self.child_spawner

    def load_child_class(self, state):
        # Subclasses must arrange for correct child_class setting from load_state
        pass

    def load_state(self, state):
        super().load_state(state)
        self.load_child_class(state)
        self.child_config.update(state.get('child_conf', {}))
        self.child_state = state.get('child_state', {})
        self.construct_child()

    def get_state(self):
        state = super().get_state()
        state['child_conf'] = self.child_config
        if self.child_spawner:
            self.child_state = state['child_state'] = self.child_spawner.get_state()
        return state

    def clear_state(self):
        super().clear_state()
        if self.child_spawner:
            self.child_spawner.clear_state()
        self.child_state = {}
        self.child_spawner = None

    # proxy functions for start/poll/stop
    # pass back the child's Future, or create a dummy if needed

    def start(self):
        if not self.child_spawner:
            self.construct_child()
        return self.child_spawner.start()

    def stop(self, now=False):
        if self.child_spawner:
            return self.child_spawner.stop(now)
        else:
            return _yield_val()

    def poll(self):
        if self.child_spawner:
            return self.child_spawner.poll()
        else:
            return _yield_val(1)

class ProfilesSpawner(WrapSpawner):

    """ProfilesSpawner - leverages the Spawner options form feature to allow user-driven
        configuration of Spawner classes while permitting:
        1) configuration of Spawner classes that don't natively implement options_form
        2) administrator control of allowed configuration changes
        3) runtime choice of which Spawner backend to launch
    """

    profiles = List(
        trait = Tuple( Unicode(), Unicode(), Type(Spawner), Dict() ),
        default_value = [ ( 'Local Notebook Server', 'local', LocalProcessSpawner,
                            {'start_timeout': 15, 'http_timeout': 10} ) ],
        minlen = 1,
        config = True,
        help = """List of profiles to offer for selection. Signature is:
            List(Tuple( Unicode, Unicode, Type(Spawner), Dict )) corresponding to
            profile display name, unique key, Spawner class, dictionary of spawner config options.

            The first three values will be exposed in the input_template as {display}, {key}, and {type}"""
        )

    child_profile = Unicode()

    form_template = Unicode(
        """<label for="profile">Select a job profile:</label>
        <select class="form-control" name="profile" required autofocus>
        {input_template}
        </select>
        """,
        config = True,
        help = """Template to use to construct options_form text. {input_template} is replaced with
            the result of formatting input_template against each item in the profiles list."""
        )

    first_template = Unicode('selected',
        config=True,
        help="Text to substitute as {first} in input_template"
        )

    input_template = Unicode("""
        <option value="{key}" {first}>{display}</option>""",
        config = True,
        help = """Template to construct {input_template} in form_template. This text will be formatted
            against each item in the profiles list, in order, using the following key names:
            ( display, key, type ) for the first three items in the tuple, and additionally
            first = "checked" (taken from first_template) for the first item in the list, so that
            the first item starts selected."""
        )

    options_form = Unicode()

    def _options_form_default(self):
        temp_keys = [ dict(display=p[0], key=p[1], type=p[2], first='') for p in self.profiles ]
        temp_keys[0]['first'] = self.first_template
        text = ''.join([ self.input_template.format(**tk) for tk in temp_keys ])
        return self.form_template.format(input_template=text)

    def options_from_form(self, formdata):
        # Default to first profile if somehow none is provided
        d = dict(profile=formdata.get('profile', [self.profiles[0][1]])[0])
        d['dyn_input'] = input = formdata.get('input')[0]
        for k,v in formdata.items():
            if k[:4] == "dyn_":
                a,input_,tag = k.split("_dyn_")
                if input_==input:
                    d["dyn_"+tag] = v[0]  
        print("options_from_form=/%s/" % (pprint.pformat(d)))
        return d

    # load/get/clear : save/restore child_profile (and on load, use it to update child class/config)

    def select_profile(self, profile):
        # Select matching profile, or do nothing (leaving previous or default config in place)
        for p in self.profiles:
            if p[1] == profile:
                self.child_class = kslhub.batchspawner.SlurmSpawner
                self.child_config = p[3]
                self.child_config['input'] = self.child_input                
                break

    def construct_child(self):
        self.child_profile = self.user_options.get('profile', "")
        self.child_input = self.user_options.get('input', "")
        self.select_profile(self.child_profile)
        for k,v in self.user_options.items():
            if k[:4] == "dyn_":
                self.child_config[k] = v
        super().construct_child()

    def load_child_class(self, state):
        try:
            self.child_profile = state['profile']
        except KeyError:
            self.child_profile = ''
        self.select_profile(self.child_profile)

    def get_state(self):
        state = super().get_state()
        state['profile'] = self.child_profile
        return state

    def clear_state(self):
        super().clear_state()
        self.child_profile = ''

# vim: set ai expandtab softtabstop=4:




#!/bin/env/python

import argparse
import os
import sys
from ._version import version_info, __version__


DEBUG = False

cmd_line = (" ".join(sys.argv))

if 'KSLHUB_PARAMS' in os.environ.keys():
  cmd_line = clean_line(os.environ['KSLHUB_PARAMS']) + " " + cmd_line



class kslhub_frontend():

  def __init__(self):

    self.parser = argparse.ArgumentParser(conflict_handler='resolve')
     
    
    self.parser.add_argument('--debug', action="count", default=0, help=argparse.SUPPRESS)
    self.parser.add_argument('--info', action="count", default=0, help=argparse.SUPPRESS)

    self.parser.add_argument('--init', action="store_true", default=False,
                             help="Initialize the hub")

    self.parser.add_argument('--jupyter-config', action="store_true", default=False,
                             help="complete some jupyter notebook initialization")

    self.parser.add_argument('--start', action="store_true", default=False,
                             help="start the hub")

    self.parser.add_argument("-v", '--version', action="store_true", default=False,
                             help="show hub current version ")


    self.parser.add_argument('--generate-config', action="store_true", default=False,
                             help="generate a default config file")

    self.parser.add_argument('--generate-job-templates', action="store_true", default=False,
                             help="generate a set of default job template files")

    self.parser.add_argument("-f", "--config", type=str, help='hub configuration file',
                             default= "default")
    
    self.parser.add_argument("--port", type=int, help='port of the hub web interface (9000 per default)',
                             default=9000)
    self.parser.add_argument("--ip", type=int, help='address of the hub web interface (0.0.0.0 per default)',
                             default=9000)

    # self.parser.add_argument("--hub-name", type=str, help='Name of the hub', default='KSLHUB')

    # self.parser.add_argument("--greeting-message", type=str,
    #                          help='Message to be shown on the welcome page of the Hub', default='KSLHUB')

    # self.parser.add_argument("--job-templates-dir", type=str, help='directory where template jobs live',
    #                          default='job_templates')

    self.args = self.parser.parse_args()




def start():
    #main()
    K = kslhub_frontend()
    args = ""

    
    if K.args.version:
       print(version_info.__version__)
       sys.exit(0)
      
    if not(os.path.exists("jobs")):
      cmd = """
         mkdir -p jobs logs runtime
         chmod 777 jobs logs runtime
      """
      os.system(cmd)
      print("creating working directory jobs, runtime and logs")

    exit_after_generate_job_templates = False
    if K.args.generate_job_templates:
      exit_after_generate_job_templates = True
      

    if K.args.generate_job_templates or not(os.path.exists("job_templates")):
      produced_template_dir = "./job_templates"
      if os.path.exists(produced_template_dir):
        print(("\nERROR: default template job directory %s already exists!\n" +\
              "\tI will not dare overwrite it...\n" + \
              "\tplease rename it or move it before generating again the default files\n" ) % \
              produced_template_dir)
        sys.exit(-1)
      cmd = "cp -r %s/share/kslhub/job_templates ./"  % (sys.base_prefix)
      os.system(cmd)
      print("Writing default template to: %s " % produced_template_dir)
      if exit_after_generate_job_templates:
        sys.exit(0)
      
    
    if K.args.jupyter_config:
      cmd = """
         # installing NERSC slurm magic  kernel
         mkdir -p /tmp/kslhub_initialization/BUILD
         cd /tmp/kslhub_initialization/BUILD
         git clone https://github.com/NERSC/slurm-magic.git
         cd slurm-magic
         python setup.py install

         cd ..
         jupyter-kernelspec install slurm-magic --sys-prefix

         # configuring the extension 
         jupyter contrib nbextension install  --sys-prefix
         jupyter nbextensions_configurator enable --sys-prefix
         jupyter nbextension enable codefolding/main
         jupyter nbextension enable --py --sys-prefix widgetsnbextension


         jupyter labextension install @jupyter-widgets/jupyterlab-manager
         jupyter labextension install  @jupyterlab/hub-extension
         jupyter labextension install  @jupyter-widgets/jupyterlab-manager   jupyter-matplotlib

         cd /tmp
         rm -rf /tmp/kslhub_initialization
      """
      os.system(cmd)
      print("kslhub is now configured")

      
    elif K.args.generate_config:
      produced_config_dir = "./kslhub_config.py"
      if os.path.exists(produced_config_dir):
        print(("\nERROR: default configuration file %s already exists!\n" +\
              "\tI will not dare overwrite it...\n" + \
              "\tplease rename it or move it before generating again the default file\n" ) % \
              produced_config_dir)
        sys.exit(-1)
      os.system("cp %s/share/kslhub/config/config.py %s"  % (sys.base_prefix, produced_config_dir))
      print("Writing default config to: %s " % produced_config_dir)
      sys.exit(0)
      
    else: 
      if K.args.config:
        if not(os.path.exists(K.args.config)):
          if os.path.exists("%s/share/kslhub/config/%s.py" % (sys.base_prefix,K.args.config)):
            K.args.config = "%s/share/kslhub/config/%s.py" % (sys.base_prefix,K.args.config)
          else:
            print("\nERROR: configuration file %s does not exist!" % K.args.config)
            print("\t- either give a python config file path")
            print("\t- or choose one of the configuration available in the directory %s/share/kslhub/config/: " % sys.base_prefix)
            print("\t\t",end="")
            for f in os.listdir("%s/share/kslhub/config/" % sys.base_prefix):
              if f.find("__pycache__")>-1 or f[-1]=="~":
                continue
              print("%s, " % f.replace(".py",""),end="")
            print("\n")
            sys.exit(-1)
            

        args = args + "-f %s" % K.args.config
      cmd = "export KSLHUB_ROOT=%s; jupyterhub %s" % (os.getcwd(),args)
      print("should start the hub with command : /%s/" % cmd)
      os.system(cmd)

    # else:
    #    print("\nChoose at least one the following options:\n\t--init to complete the installation of kslhub\n\t--start ro start the hub"+\
    #          "\n\t--generate-config to generate an example of configuration file\n\t--help to have the following message\n\n")
    #    K.parser.parse_args(['-h'])

  
if __name__ == "__main__":
    K = kslhub_frontend()




import os
import socket
import sys

s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
s.connect(("8.8.8.8", 80))
current_host=s.getsockname()[0]

# force your own interface here... interface has to be reachable from a compute node
# current_host="10.129.35.32"

hub_port = 8000
proxy_port = 9800


print('current_host = %s' % current_host)


c.Authenticator.admin_users = {"kortass"}

c.JupyterHub.allow_named_servers = False

c.JupyterHub.hub_ip = current_host
c.JupyterHub.hub_port = hub_port
c.JupyterHub.bind_url = 'http://%s:%s' % (c.JupyterHub.hub_ip,c.JupyterHub.hub_port)
c.JupyterHub.hub_bind_url = 'http://%s:%s' % (current_host,proxy_port)
c.ConfigurableHTTPProxy.api_url = 'http://%s:%s' % (current_host,proxy_port)

c.Spawner.debug  = False

c.Authenticator.otp_required = False 
c.Authenticator.host = 'localhost'

c.Authenticator.hub_greeting_message = "Welcome to KSL Hub @ SK!"
c.Authenticator.hub_name = "SK"
c.Authenticator.job_template_dir = "./job_templates"


c.JupyterHub.authenticator_class = 'kslhub.otp_authenticator.SshUserAuthenticator'
c.JupyterHub.template_paths = [ "%s/share/kslhub/templates" % sys.base_prefix ]



c.Spawner.cmd = ['jupyter-labhub']

runtime_dir = os.path.join('/scratch/tmp')


c.JupyterHub.spawner_class = 'kslhub.wrapspawner.ProfilesSpawner'
c.Spawner.http_timeout = 120
c.BatchSpawnerBase.req_host = 'localhost'
c.BatchSpawnerBase.req_runtime = '12:00:00'
c.BatchSpawnerBase.hub_port = hub_port
c.BatchSpawnerBase.hub_ip = current_host
c.BatchSpawnerBase.proxy_port = proxy_port
c.BatchSpawnerBase.proxy_ip = current_host

c.InteractiveShellApp.extensions = ['slurm_magic']
c.Spawner.http_timeout = 300
c.Spawner.notebook_dir = '~/NOTEBOOKS' 
c.Spawner.start_timeout = 300



c.ProfilesSpawner.profiles = [
   ('Slurm - 1 Node - debug queue - 1 minutes', 'debug01', 'kslhub.batchspawner.SlurmSpawner',
      dict(req_nprocs='1', req_partition='debug', req_runtime='0:01:00')),
   ('Slurm - 1 Node - debug queue - 2 minutes', 'debug02', 'kslhub.batchspawner.SlurmSpawner',
      dict(req_nprocs='1', req_partition='debug', req_runtime='0:02:00')),
   ('Slurm - 1 Node - debug queue - 5 minutes', 'debug05', 'kslhub.batchspawner.SlurmSpawner',
      dict(req_nprocs='1', req_partition='debug', req_runtime='0:05:00')),
   ('Slurm - 1 Node - debug queue - 30 minutes', 'debug30', 'kslhub.batchspawner.SlurmSpawner',
      dict(req_nprocs='1', req_partition='debug', req_runtime='0:30:00')),
   ( "Local server", 'local', 'jupyterhub.spawner.LocalProcessSpawner', {'ip':'0.0.0.0'} ),
   ]



import re


def build_time(tag,fields,case):
    tag_text = fields.pop(0)
    s = """
          <tr class='additional case___CASE__'>
            <td>
              <label class="control-label">__TAG_TEXT__</label>&nbsp &nbsp;
            </td>
            <td >
              <input id="__CASE__-__TAG__-hours"   size="2" value="0"  type="tel" 
                     onChange="$('#__CASE__-__TAG__-whole').val($('#__CASE__-__TAG__-hours').val()+':'+$('#__CASE__-__TAG__-minutes').val()+':00')" > h
              <input id="__CASE__-__TAG__-minutes" size="2" value="30" type="tel"
                     onChange="$('#__CASE__-__TAG__-whole').val($('#__CASE__-__TAG__-hours').val()+':'+$('#__CASE__-__TAG__-minutes').val()+':00')" > m
              <input id="__CASE__-__TAG__-whole" name="dyn_dyn___CASE___dyn___TAG__" size="5" value="00:30:00"  type="hidden">
            </td>
          </tr>

    """
    s = s.replace('__CASE__',case).replace('__TAG_TEXT__',tag_text).replace('__TAG__',tag)
    return s

def build_select(tag,fields,case):
    tag_text = fields.pop(0)
    s = """
          <tr class='additional case_%s'>
            <td>
              <label class="control-label">%s</label>&nbsp &nbsp;
            </td>
            <td>
              <select name="dyn_dyn_%s_dyn_%s" class="form-control dynamic jobform">
    """ % (case,tag_text,case,tag)

    options = fields.pop(0).split(",")
    for o in options:
        s = s + """\t\t<option value="%s">%s</option>
        """ % (o,o)
    s = s + """
              </select>
            </td>
          </tr>
    """
    return s

def build_input(tag,fields,case):
    tag_text = fields.pop(0)
    tag_type = "text"
    tag_default_value = ""
    if len(fields):
        tag_type = fields.pop(0)
        if len(fields):
            tag_default_value = fields.pop(0)
    s = """
          <tr class='additional case_%s'>
            <td>
              <label class="control-label">%s</label>&nbsp &nbsp;
            </td>
            <td>
              <input name="dyn_dyn_%s_dyn_%s" type="%s" value="%s">


    """ % (case,tag_text,case,tag,tag_type,tag_default_value)


    s = s + """
            </td>
          </tr>
    """
    return s


def build_gui_from_template(job_file,case):

    job_file_content = "".join(open(job_file,"r").readlines())

    # wrong way to parse: couldn't catch __xxx__/__yyy__
    #    xxx__/__yyy was detected as variable
    # variables =  [ x[2:-2] for x in re.findall('__.+__', job_file_content)]
    
    v = job_file_content.split("__")
    variables=[ v[i*2+1] for i in range(len(v)//2)  ]

    tags = {}

    l = ""

    for v in variables:
        print(v)

        fields = v.split(";")

        tag = fields.pop(0)

        if len(fields):

            web_type = fields.pop(0)

            if web_type == "select":

                s = build_select(tag,fields,case)
                l = l + s

            elif web_type == "time":
        
                s = build_time(tag,fields,case)
                l = l +s

            elif web_type == "input":
        
                s = build_input(tag,fields,case)
                l = l +s
    return l

                
if __name__ == "__main__":
    job_file = "/home/kortass/JUPYTER/job_templates/sk/00 - lab.template"

    l = build_gui_from_template(job_file)
    print(l)

    
    




"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import glob
import os
import sys
import kslhub

here = os.path.abspath(os.path.dirname(__file__))


# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


pjoin = os.path.join

share_jupyterhub = pjoin(here, 'jupyterhub','share', 'jupyterhub')
static = pjoin(share_jupyterhub, 'static')

is_repo = os.path.exists(pjoin(here, '.git'))

# ---------------------------------------------------------------------------
# Build basic package data, etc.
# ---------------------------------------------------------------------------


def get_data_files():
    """Get data files in share/jupyter"""

    data_files = []
    ntrim = len(here + os.path.sep)

    for (d, dirs, filenames) in os.walk(pjoin('templates')):
       data_files.append(('share/kslhub/templates/', [pjoin(d, f) for f in filenames]))

    for (d, dirs, filenames) in os.walk(pjoin('job_templates')):
       data_files.append(('share/kslhub/job_templates/', [pjoin(d, f) for f in filenames]))

    for (d, dirs, filenames) in os.walk(pjoin('config')):
       data_files.append(('share/kslhub/config/', [pjoin(d, f) for f in filenames]))

    site_packages_dir = "lib/python%d.%d/site-packages" % (sys.version_info[0],sys.version_info[1])
    
    data_files = data_files + \
                 [ ("%s/jupyterhub" % site_packages_dir,['jupyterhub/jupyterhub/orm.py',
                                                         'jupyterhub/jupyterhub/spawner.py' ]), \
                   ("%s/jupyterhub/oauth" % site_packages_dir, ['jupyterhub/jupyterhub/oauth/provider.py']), \
                   ('%s/jupyterhub/handlers' % site_packages_dir, ['jupyterhub/jupyterhub/handlers/login.py',
                                            'jupyterhub/jupyterhub/handlers/pages.py'])]
    return data_files



# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='kslhub',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=kslhub.__version__.replace(".dev",""),  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='klshub - a datahub based on jupyterhub ',  # Required

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='http://kslhub.readthedocs.io',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='Samuel KORTAS',  # Optional
    maintainer ='Samuel KORTAS',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='samuel.kortas@kaust.edu.sa',  # Optional
    maintainer_email='samuel.kortas@kaust.edu.sa',  # Optional

    # source_label and source_url are described in draft PEP426. they are
    # inserted into the package's metadata, which can be retrieved by using
    # `py-info <package>` on any installed package. the contents are not
    # validated to conform to any spec other than being a string
    # source_label = "eb745f9bf1e0c6d749d5206f66cb41555fbb9013",

    # same with source_url, it's only in the metadata
    # source_url = "https://github.com/samkos/kslhub/commit/eb745f9bf1e0c6d749d5206f66cb41555fbb9013",

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        'Environment :: Console',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        #'Programming Language :: Python :: 2',
        #'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'Operating System :: POSIX',
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Clustering",
        "Topic :: System :: Distributed Computing",
        "Topic :: Utilities"

    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='scheduler extension workflow parametric',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['contrib', 'docs', 'tests','jupyterhub']),  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['cython',
                      'notebook',
                      'pandas',
                      'matplotlib',
                      'sympy  ',
                      'bash_kernel',
                      'yapf',
                      'jupyter_contrib_nbextensions',
                      'jupyter_nbextensions_configurator ',
                      'ipywidgets',
                      'dask',
                      'toolz',
                      'cloudpickle ',
                      'distributed',
                      'dockerspawner',
                      'ipyparallel',
                      'six>-1.11.0',
                      'jupyter_contrib_nbextensions',
                      'alembic',
                      'async_generator>=1.8',
                      'entrypoints',
                      'traitlets>=4.3.2',
                      'tornado>=5.0',
                      'jinja2',
                      'oauthlib>=2,<3',
                      'pamela',
                      'paramiko',
                      'python-dateutil',
                      'SQLAlchemy>=1.1',
                      'requests',
                      'prometheus_client>=0.0.21',
                      'jupyterhub==0.9.4',
                      'jupyterlab',
                      'certipy>=0.1.2'],  # Optional

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    # extras_require={  # Optional
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    #package_data = get_package_data(),

    python_requires='>=3.6',
  
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files= get_data_files(),  # Optional

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={  # Optional
          'console_scripts': [
             'kslhub=kslhub.kslhub_frontend:start',
         ],
     },
)




import os
from jupyterhub.handlers import BaseHandler
from jupyterhub.auth import Authenticator
from jupyterhub.auth import LocalAuthenticator
from jupyterhub.utils import url_path_join
from tornado import gen, web
from traitlets import Unicode, Bool, Integer

import pexpect
import pexpect.pxssh as pxssh
import subprocess
import sys, os
import getpass


from .ssh_authenticator import session_manager

def otp_validate(hostname,port,user,pw,otp):

    print("Username: %s" % user)
    print("Hostname: %s:%s" % (hostname,port))
    print("OTP is: %s" % otp)

    ssh_newkey = 'Are you sure you want to continue connecting'
    # my ssh command line
    p=pexpect.spawn('ssh -p %s %s@%s echo hostname:`hostname`' % (port,user,hostname))

    password_already_sent = False
    otp_already_sent = False

    authenticated = False
    
    while True:
      i=p.expect([ssh_newkey,'assword:',"One-time .*:","Verification .*:","hostname .*$", pexpect.EOF])
      if i==0:
        print("I say yes")
        p.sendline('yes')
      if i==1:
        if password_already_sent:
          authenticated = False
          print("I aleady gave password")
          break
        print("I give password")
        p.sendline(pw)
        password_already_sent = True
      if i==2 or i==3:
        if otp_already_sent:
          authenticated = False
          print("I aleady gave otp")
          break
        print("I give otp")
        p.sendline(otp)
        otp_already_sent = True
      if i==4:
        authenticated = True
        break
      elif i==5:
        print("I either got key or connection timeout")
        break
        pass

    print(p.before) # print out the result
    if not(authenticated):
      authenticated  =  str(p.before).find('hostname')>-1
    p.close()
    
    if authenticated:
      print("SUCESS login %s on %s" % (user,hostname))
    else:
      print("FAILED login %s on %s" % (user,hostname))
      
    print(" ")

    return authenticated

slurm_machine = session_manager()
    
class SshUserLoginHandler(BaseHandler):

    def get(self):
        header_name = self.authenticator.header_name
        ssh_user = self.request.headers.get(header_name, "")
        if ssh_user == "":
            raise web.HTTPError(401)

        user = self.user_from_username(ssh_user)
        self.set_login_cookie(user)
        next_url = self.get_next_url(user)
        self.redirect(next_url)


class SshUserAuthenticator(Authenticator):
    """
    Accept the authenticated user name from the SSH_USER HTTP header.
    """
    
    hub_greeting_message = Unicode("DataHub Sign in", config=True,
        help="""
        Greeting message to print in the login box
        """
    ).tag(config=True)
    
    otp_required = Bool(False, config=True,
        help="""Ask for an OTP password if True
        """,
    )

    otp_authenticator_host = Unicode("localhost", config=True,
        help="""hostname of where to validate OTP password if otp_required set to True
        """,
    )

    otp_authenticator_port = Integer(22, config=True,
        help="""port of hostname of where to validate OTP password if otp_required set to True
        """,
    )

    hub_name = Unicode("DataHub", config=True,
        help="""
        Greeting message to print in the login box
        """
    ).tag(config=True)
    
    job_template_dir = Unicode("./job_templates", config=True,
        help="""
        job templates directory
        """
    ).tag(config=True)


    
    @gen.coroutine
    def authenticate(self, handler, data):

        if handler.config.Authenticator.otp_required:
            hostname = handler.config.Authenticator.host
            port = handler.config.Authenticator.port
            login_ok = otp_validate(hostname,
                                    port,
                                    data['username'],
                                    data['password'],
                                    data['otp'])
            self.log.info("attempt of authentication of %s@%s:%s with otp -> result = %s" % \
                     (data['username'],hostname,port,login_ok))
            if (not(login_ok)):
                return None
                
        machine = handler.config.Authenticator.host
        # self.log.info("hello")
        if slurm_machine.open(machine,data['username'],data['password']):
            return data['username']
        return None

class SshUserLocalAuthenticator(LocalAuthenticator):
    """
    Accept the authenticated user name from the SSH_USER HTTP header.
    Derived from LocalAuthenticator for use of features such as adding
    local accounts through the admin interface.
    """
    header_name = Unicode(
        default_value='SSH_USER',
        config=True,
        help="""HTTP header to inspect for the authenticated username.""")

    def get_handlers(self, app):
        return [
            (r'/login', SshUserLoginHandler),
        ]

    @gen.coroutine
    def authenticate(self, *args):
        raise NotImplementedError()




"""KSLHub version info"""

# Copyright (c) KslHub Team.
# Distributed under the terms of the Modified BSD License.

version_info = (
    0,
    0,
    17,
    "",  # release (b1, rc1, or "" for final or dev)
    "dev",  # dev or nothing
)

# pep 440 version: no dot before beta/rc, but before .dev
# 0.1.0rc1
# 0.1.0a1
# 0.1.0b1.dev
# 0.1.0.dev

__version__ = ".".join(map(str, version_info[:3])) + ".".join(version_info[3:])
